"use strict";

var fs = require('fs');

var Gpio = module.exports = function Gpio() {
    var mtsIoDir = "/sys/devices/platform/mts-io";
    var ac1Dir = mtsIoDir.concat("/ap1");
    var ac2Dir = mtsIoDir.concat("/ap2");
    // auto detect port
    var acDir = mtsIoDir.concat("/gpiob");

    /* digital output ports ***************************************************/
    var do0 = "dout0";
    var do1 = "dout1";
    var do2 = "dout2";
    var do3 = "dout3";

    var ac1Do0 = ac1Dir.concat("/", do0);
    var ac1Do1 = ac1Dir.concat("/", do1);
    var ac1Do2 = ac1Dir.concat("/", do2);
    var ac1Do3 = ac1Dir.concat("/", do3);

    var ac2Do0 = ac2Dir.concat("/", do0);
    var ac2Do1 = ac2Dir.concat("/", do1);
    var ac2Do2 = ac2Dir.concat("/", do2);
    var ac2Do3 = ac2Dir.concat("/", do3);

    var acDo0 = acDir.concat("/", do0);
    var acDo1 = acDir.concat("/", do1);
    var acDo2 = acDir.concat("/", do2);
    var acDo3 = acDir.concat("/", do3);

    this.digitalOutputPaths = [[ac1Do0, ac2Do0],
                               [ac1Do1, ac2Do1],
                               [ac1Do2, ac2Do2],
                               [ac1Do3, ac2Do3]];

    this.autoDigitalOutputPaths = [acDo0, acDo1, acDo2, acDo3];

    /** digital input ports ***************************************************/
    var di0 = "din0";
    var di1 = "din1";
    var di2 = "din2";
    var di3 = "din3";

    var ac1Di0 = ac1Dir.concat("/", di0);
    var ac1Di1 = ac1Dir.concat("/", di1);
    var ac1Di2 = ac1Dir.concat("/", di2);
    var ac1Di3 = ac1Dir.concat("/", di3);

    var ac2Di0 = ac2Dir.concat("/", di0);
    var ac2Di1 = ac2Dir.concat("/", di1);
    var ac2Di2 = ac2Dir.concat("/", di2);
    var ac2Di3 = ac2Dir.concat("/", di3);
    this.digitalInputPaths = [[ac1Di0, ac2Di0],
                              [ac1Di1, ac2Di1],
                              [ac1Di2, ac2Di2],
                              [ac1Di3, ac2Di3]];

    // auto detected port inputs
    var acDi0 = acDir.concat("/", di0);
    var acDi1 = acDir.concat("/", di1);
    var acDi2 = acDir.concat("/", di2);
    var acDi3 = acDir.concat("/", di3);

    this.autoDigitalInputPaths = [acDi0, acDi1, acDi2, acDi3];

    /** analog input ports ****************************************************/
    var ai0 = "adc0";
    var ai1 = "adc1";
    var ai2 = "adc2";
    var ai3 = "adc3";

    var ac1Ai0 = ac1Dir.concat("/", ai0);
    var ac1Ai1 = ac1Dir.concat("/", ai1);
    var ac1Ai2 = ac1Dir.concat("/", ai2);
    var ac1Ai3 = ac1Dir.concat("/", ai3);

    var ac2Ai0 = ac2Dir.concat("/", ai0);
    var ac2Ai1 = ac2Dir.concat("/", ai1);
    var ac2Ai2 = ac2Dir.concat("/", ai2);
    var ac2Ai3 = ac2Dir.concat("/", ai3);
    this.analogInputPaths = [[ac1Ai0, ac2Ai0],
                             [ac1Ai1, ac2Ai1],
                             [ac1Ai2, ac2Ai2],
                             [ac1Ai3, ac2Ai3]];

    // auto detected port inputs
    var acAi0 = acDir.concat("/", ai0);
    var acAi1 = acDir.concat("/", ai1);
    var acAi2 = acDir.concat("/", ai2);
    var acAi3 = acDir.concat("/", ai3);
    this.autoAnalogInputPaths = [acAi0, acAi1, acAi2, acAi3];
}

/**
 * Calculates the sample interval to use based on the chosen rate in Hertz.
 *
 * rateHz - the desired rate in Hertz (times per second)
 * returns - the milliseconds to sleep between samples to achieve that rate
 */
Gpio.prototype.calculateSampleInterval = function (rateHz) {
    if (rateHz == null || rateHz == 0) {
        return 0;
    } else {
        return 1000/rateHz;
    }
}

/**
 * Checks to see if the file exists
 * path - the path of the file to check
 * returns true if it exists, otherwise false
 */
Gpio.prototype.fileExists = function (path) {
    var exists = path != null && fs.existsSync(path);
    if (!exists) {
        console.error("Path does not exist: %s. Unable to perform operation on file.", path);
    }
    return exists;
}

/**
 * Synchronously writes the value in a file.
 * path - the path of the file to write
 * returns true if successful, otherwise false
 */
Gpio.prototype.writeValue = function(path, value) {
    if (this.fileExists(path)) {
        fs.writeFileSync(path, value);
        return true;
    }
    return false;
}

/**
 * Synchronously reads the value in a file and returns it.
 * path - the path of the file to read
 * returns the value found in the file or null if there is a problem
 */
Gpio.prototype.readValue = function(path) {
    if (this.fileExists(path)) {
        return fs.readFileSync(path, "ascii");
    }
    return null;
}
Gpio.prototype.isValidGpioNumber = function (gpioNumber) {
    var result = (gpioNumber != null) && (gpioNumber == 0 || gpioNumber == 1 || gpioNumber == 2 || gpioNumber == 3);
    if (!result) {
        console.error("Invalid GPIO number: %s", gpioNumber);
    }
    return result;
}

Gpio.prototype.isValidAccessoryCardPort = function (accessoryCardPort) {
    var result = (accessoryCardPort != undefined)
        && (accessoryCardPort != null)
        && (accessoryCardPort == 1 || accessoryCardPort == 2);
    if (!result) {
        console.error("Invalid accessory card port: %s", accessoryCardPort);
    }
    return result;
}

Gpio.prototype.createInt = function (value) {
   if (value == null || value == undefined) {
      return null;    
   } else {
      return parseInt(value);
   }
}

Gpio.prototype.createFloat = function (value) {
   if (value == null || value == undefined) {
      return null;    
   } else {
      return parseFloat(value);
   }
}

/** analog ********************************************************************/

Gpio.prototype.detectAnalogInputPath = function (gpioNumber) {
    if (!this.isValidGpioNumber(gpioNumber)) {
        return null;
    }
    return this.autoAnalogInputPaths[gpioNumber];
}

/**
 * Gets the analog input GPIO sysfs path for the given accessory card and
 * GPIO number.
 * accessoryCardPort - the port of the accessory card
 * gpioNumber - the number associated with the GPIO
 */
Gpio.prototype.getAnalogInputPath = function (accessoryCardPort, gpioNumber) {
    if (!this.isValidAccessoryCardPort(accessoryCardPort)) {
        return null;
    }
    if (!this.isValidGpioNumber(gpioNumber)) {
        return null;
    }
    return this.analogInputPaths[gpioNumber][accessoryCardPort - 1];
}

Gpio.prototype.readAnalogValue = function (accessoryCardPort, gpioNumber) {
    var path = this.getAnalogInputPath(accessoryCardPort, gpioNumber);
    return this.createFloat(this.readValue(path));
}

Gpio.prototype.detectAndReadAnalogValue = function (gpioNumber) {
    var path = this.detectAnalogInputPath(gpioNumber);
    return this.createFloat(this.readValue(path));
}

/**
 * Runs the state changed filter.
 * enabled - Is the filter enabled
 * relationalOperator - The relational operator to use
 * value - The new value
 * thresholdValue - the value to compare the new value to
 * returns 
 *    true if enabled is false
 *    true if enabled is true and this expression is true: value (relationalOperator) thresholdValue
 *    false, otherwise
 */
Gpio.prototype.valueRangeFilterPass = function (enabled, value, relationalOperator, thresholdValue) {
   if (!enabled) {
      return true;   
   }
   switch (relationalOperator) {
      case ">=":
         return value >= thresholdValue;
      case ">":
         return value > thresholdValue;
      case "<=":
         return value <= thresholdValue;
      case "<":
         return value < thresholdValue;
   }
	return false;
}

/** digital *******************************************************************/

/**
 * Detects the accessory card port and returns the associated path to the
 * requested GPIO pin.  If there are two accessory cards of the same type on the
 * system, port 1 is always used.
 *
 * gpioNumber - the number associated with the GPIO pin
 * returns the digital output path for the given GPIO number or null if the
 * number is invalid
 */
Gpio.prototype.detectDigitalOutputPath = function (gpioNumber) {
    if (!this.isValidGpioNumber(gpioNumber)) {
        return null;
    }
    return this.autoDigitalOutputPaths[gpioNumber];
}

/**
 * Detects the accessory card port and returns the associated path to the
 * requested GPIO pin.  If there are two accessory cards of the same type on the
 * system, port 1 is always used.
 *
 * gpioNumber- the number associated with the GPIO pin
 * returns the digital input path for the given GPIO number or null if the
 * number is invalid
 */
Gpio.prototype.detectDigitalInputPath = function (gpioNumber) {
    if (!this.isValidGpioNumber(gpioNumber)) {
        return null;
    }
    return this.autoDigitalInputPaths[gpioNumber];
}

/**
 * Gets the digital output GPIO sysfs path for the given accessory card and
 * GPIO number.
 * accessoryCardPort - the port of the accessory card
 * gpioNumber - the number associated with the GPIO
 */
Gpio.prototype.getDigitalOutputPath = function (accessoryCardPort, gpioNumber) {
    if (!this.isValidAccessoryCardPort(accessoryCardPort)) {
        return null;
    }
    if (!this.isValidGpioNumber(gpioNumber)) {
        return null;
    }
    return this.digitalOutputPaths[gpioNumber][accessoryCardPort - 1];
}

/**
 * Gets the digital input GPIO sysfs path for the given accessory card and
 * GPIO number.
 * accessoryCardPort - the port of the accessory card
 * gpioNumber - the number associated with the GPIO
 */
Gpio.prototype.getDigitalInputPath = function (accessoryCardPort, gpioNumber) {
    if (!this.isValidAccessoryCardPort(accessoryCardPort)) {
        return null;
    }
    if (!this.isValidGpioNumber(gpioNumber)) {
        return null;
    }
    return this.digitalInputPaths[gpioNumber][accessoryCardPort - 1];
}

/**
 * Synchronously writes the digital value to a file
 * accessoryCardPort - the accessory card port number
 * gpioNumber - the GPIO pin number
 * value - the value to write, this must be 0 or 1 otherwise it will fail
 * returns true if successful, otherwise false
 */
Gpio.prototype.writeDigitalValue = function (accessoryCardPort, gpioNumber, value) {
    if (this.isDigitalValue(value)) {
        var path = this.getDigitalOutputPath(accessoryCardPort, gpioNumber);
        return this.writeValue(path, value);
    }
    return false;
}

Gpio.prototype.detectAndWriteDigitalValue = function (gpioNumber, value) {
    if (this.isDigitalValue(value)) {
        var path = this.detectDigitalOutputPath(gpioNumber);
        return this.writeValue(path, value);
    }
    return false;
}

/**
 * Synchronously reads the digital value from the file and returns it.
 * accessoryCardPort - the accessory card port number
 * gpioNumber - the GPIO pin number
 * returns the value found in the file or null if there is a problem
 */
Gpio.prototype.readDigitalValue = function (accessoryCardPort, gpioNumber) {
    var path = this.getDigitalInputPath(accessoryCardPort, gpioNumber);
    return this.createInt(this.readValue(path));
}

Gpio.prototype.detectAndReadDigitalValue = function (gpioNumber) {
    var path = this.detectDigitalInputPath(gpioNumber);
    return this.createInt(this.readValue(path));
}

/**
 * Checks to see if the given value is a valid digital value. In other words,
 * it checks to see if the value is a 1 or 0.
 * returns true if the value is a 1 or 0, otherwise false
 */
Gpio.prototype.isDigitalValue = function (value) {

    if (value == null) {
        console.error("Incorrect digital value provided: %d", value);
        return false;
    }
    if (value == 0 || value == 1) {
        return true;
    }
    console.error("Invalid digital value given: %d", value);
    return false;
}

/**
 * Runs the state changed filter.
 * enabled - Is the filter enabled
 * previousValue - The previous value
 * value - The new value
 * returns true if enabled is false or if previousValue is different from value
 */
Gpio.prototype.stateFilterValueChanged = function (enabled, previousValue, value) {
	return !enabled || (previousValue != value);
}
