"use strict";
var exec = require('child_process').exec;
var fs = require('fs');
var EventEmitter = require("events").EventEmitter;
var serialport = require("serialport");

var Serial = module.exports = function Serial(accessoryport, mode, callback) {
  this.mtsiosysfs = new MtsIoSysfs(accessoryport, mode);
  this.mtsiosysfs.execMtsIoSysfs(callback);
  this.eventEmitter = new EventEmitter();
}

Serial.prototype.getEE = function() {
  return this.eventEmitter;
}

Serial.prototype.byteLengthParser = function(length) {
  var data = new Buffer(0);
  return function(emitter, buffer) {
    data = Buffer.concat([data, buffer]);
    var start = 0;
    var end = length;
    while (end <= data.length) {
      var out = new Buffer(length);
      // The end is goofy in Javascript.  Basically, end+1 for all operations
      data.copy(out, 0, start, end+1);
      emitter.emit('data', out);
      start = end;
      end = length + end;
    }
    if (data.length % length == 0) {
      data = new Buffer(0);
    } else {
      data = data.slice(start);
    }
  };
}

Serial.prototype.characterLengthParser = function(length, encoding) {
  var stringBuffer = "";
  return function(emitter, buffer) {
    stringBuffer = stringBuffer + buffer.toString(encoding);
    var start = 0;
    var end = length;
    while (end <= stringBuffer.length) {
      var out = new Buffer(stringBuffer.substring(start, end), encoding);
      emitter.emit('data', out);
      start = end;
      end = length + end;
    }
    if (stringBuffer.length % length == 0) {
      stringBuffer = "";
    } else {
      stringBuffer = stringBuffer.substring(start);
    }
  };
}
Serial.prototype.characterLengthParserAscii = function(length) {
  return this.characterLengthParser(length, "ascii");
}

Serial.prototype.characterLengthParserUtf8 = function(length) {
  return this.characterLengthParser(length, "utf8");
}

Serial.prototype.createReader = function(baud, databits, parity, stopbits, xon,
  xoff, parser) {
  console.log("Reader is using tty: %s", this.mtsiosysfs.getTtyPath());
  var self = this;
  this.port = new serialport.SerialPort(this.mtsiosysfs.getTtyPath(), {
    baudrate: baud,
    databits: databits,
    parity: parity,
    stopbits: stopbits,
    parser: parser,
    xon: xon,
    xoff: xoff
  }, true, function(error) {
    if (error) {
      console.error("Error opening serial port for reading.  %s", error);
    } else {
      self.mtsiosysfs.execSetRs485(function() {
        // do nothing
      });
    }
  });
}

Serial.prototype.createWriter = function(baud, databits, parity, stopbits, xon,
    xoff) {
  console.log(
      "Creating serial writer, tty: %s, baud: %d, databits: %d, parity: %s, stopbits: %d, xon: %s, xoff: %s",
      this.mtsiosysfs.getTtyPath(),
      baud, databits, parity, stopbits, xon, xoff);
  var obj = this;
  this.port = new serialport.SerialPort(this.mtsiosysfs.getTtyPath(), {
    baudrate: baud,
    databits: databits,
    parity: parity,
    stopbits: stopbits,
    xon: xon,
    xoff: xoff
  }, false);
}

Serial.prototype.createRaw = function(baud, databits, parity, stopbits,
    xon, xoff) {
  console.log(
      "Creating raw serial reader, baud: %d, databits: %d, parity: %s, stopbits: %d, xon: %s, xoff: %s",
      baud, databits, parity, stopbits, xon, xoff);
  var obj = this;
  this.createReader(baud, databits, parity, stopbits, xon, xoff,
      serialport.parsers.raw);
  this.openBinaryReader();
}

Serial.prototype.createReadLine = function(baud, databits, parity,
    stopbits, xon, xoff, delimiter, encoding) {
  console.log(
      "Creating read line serial reader, baud: %d, databits: %d, parity: %s, stopbits: %d, xon: %s, xoff: %s, delimiter: %s, encoding: %s",
      baud, databits, parity, stopbits, xon, xoff, delimiter, encoding);
  var obj = this;
  this.createReader(baud, databits, parity, stopbits, xon, xoff,
      serialport.parsers.readline(delimiter, encoding));
  this.openStringReader(encoding);
}

Serial.prototype.createByteLength = function(baud, databits, parity,
    stopbits, xon, xoff, byteLength) {
  console.log(
      "Creating byte length serial reader, baud: %d, databits: %d, parity: %s, stopbits: %d, xon: %s, xoff: %s, bytelength: %d",
      baud, databits, parity, stopbits, xon, xoff, byteLength);
  var obj = this;
  this.createReader(baud, databits, parity, stopbits, xon, xoff,
      this.byteLengthParser(byteLength));
  this.openBinaryReader();
}

Serial.prototype.createCharacterLength = function(baud, databits, parity,
    stopbits, xon, xoff, characterLength, encoding) {
  console.log(
      "Creating character length serial reader, baud: %d, databits: %d, parity: %s, stopbits: %d, xon: %s, xoff: %s, characterLength: %d, encoding: %s",
      baud, databits, parity, stopbits, xon, xoff, characterLength, encoding);
  var obj = this;
  this.createReader(baud, databits, parity, stopbits, xon, xoff,
      this.characterLengthParser(characterLength, encoding));
  this.openStringReader(encoding);
}

Serial.prototype.openStringReader = function(encoding) {
  var obj = this;
  obj.port.on("open", function () {
    console.log("Opening string reader.");
    obj.port.on('data', function(data) {
      var message = data.toString(encoding);
//          console.log('String serial data received: %s', message);
      obj.eventEmitter.emit("message", message);
    });
  });
}

Serial.prototype.openBinaryReader = function() {
  var obj = this;
  obj.port.on("open", function () {
    console.log("Opening binary reader.");
    obj.port.on('data', function(data) {
//          console.log('Binary serial data received: %s', data);
      obj.eventEmitter.emit("message", data);
    });
  });
}
/**
 * Writes the buffer to the serial port.
 * buffer - the buffer to write to the serial port
 */
Serial.prototype.writeBuffer = function(buffer, callback) {
//  console.log("Writing buffer to serial: %s.", buffer);
  var obj = this;
  if (obj.port.isOpen()) {
    obj.port.write(buffer, function(err) {
      callback(err);
    });
  } else {
    obj.port.open(function (error) {
      if (error) {
        console.error(error);
      } else {
        obj.mtsiosysfs.execSetRs485(function() {
          obj.port.write(buffer, function(err) {
            callback(err);
          });
        });
      }
    });
  }

}

/**
 * Writes the string to the serial port using ASCII encoding.
 * buffer - the buffer to write to the serial port, it may be a buffer, string,
 * or array of bytes
 */
Serial.prototype.writeAscii = function(string, callback) {
//  console.log("Writing ASCII message to serial: %s.", string);
  this.writeBuffer(new Buffer(string, "ascii"), callback);
}

/**
 * Writes the string to the serial port using UTF8 encoding.
 * buffer - the buffer to write to the serial port, it may be a buffer, string,
 * or array of bytes
 */
Serial.prototype.writeUtf8 = function(string, callback) {
//  console.log("Writing UTF8 message to serial: %s.", string);
  this.writeBuffer(new Buffer(string, "utf8"), callback);
}

Serial.prototype.close = function() {
  this.port.drain(function(err) {
    if (err) {
      console.error('Error draining serial port: ' + err);
    }
  });
  this.port.close(function(err) {
      if (err) {
        console.error('Error closing serial port: ' + err);
      }
  });
  delete this.port;
}

var MtsIoSysfs = function MtsIoSysfs(accessoryport, mode) {
  console.log("Configuring mtsiosysfs with accessory port: %s, mode: %s", accessoryport, mode);
  this.TTY1_PATH = "/dev/ttyAP1";
  this.TTY2_PATH = "/dev/ttyAP2";
  this.TTY_AUTODETECT = "";
  this.AP1_AC = "ap1";
  this.AP2_AC = "ap2";
  this.AUTODETECT_AC = "mfser";
  this.LOOPBACK_MODE = "loopback";
  this.RS422_FD_MODE = "rs422-485-full";
  this.RS485_FD_MODE = "rs422-485-full";
  this.RS485_HD_MODE = "rs485-half";
  this.RS232_MODE = "rs232";
  
  if (mode === "rs232") {
    this.setModeRs232();
  } else if (mode === "rs422fd") {
    this.setModeRs422Fd();
  } else if (mode === "rs485hd") {
    this.setModeRs485Hd();
  } else if (mode === "rs485fd") {
    this.setModeRs485Fd();
  } else if (mode === "loopback") {
    this.setModeLoopback();
  }
  if (accessoryport === "autodetect") {
    this.setAccessoryCardAutoDetect();
  } else if (parseInt(accessoryport) === 1) {
    this.setAccessoryCard1();
  } else if (parseInt(accessoryport) === 2){
    this.setAccessoryCard2();
  }
}

MtsIoSysfs.prototype.getTtyPath = function(){
  if (this.accessoryCard === this.AP1_AC) {
    return this.TTY1_PATH;
  } else if (this.accessoryCard === this.AP2_AC) {
    return this.TTY2_PATH;
  } else { // TODO, add the correct auto detect path
    return this.TTY1_PATH;
  }
}

MtsIoSysfs.prototype.setAccessoryCardAutoDetect = function(){
  this.accessoryCard = "mfser";
}
MtsIoSysfs.prototype.setAccessoryCard1 = function(){
  this.accessoryCard = this.AP1_AC;
}
MtsIoSysfs.prototype.setAccessoryCard2 = function(){
  this.accessoryCard = this.AP2_AC;
}
MtsIoSysfs.prototype.setModeLoopback = function(){
  this.mode = this.LOOPBACK_MODE;
}

MtsIoSysfs.prototype.setModeRs422Fd = function(){
  this.mode = this.RS422_FD_MODE;
}
MtsIoSysfs.prototype.setModeRs232 = function(){
  this.mode = this.RS232_MODE;
}
MtsIoSysfs.prototype.setModeRs485Hd = function(){
  this.mode = this.RS485_HD_MODE;
}
MtsIoSysfs.prototype.setModeRs485Fd = function(){
  this.mode = this.RS485_FD_MODE;
}
MtsIoSysfs.prototype.isRs485HdMode = function() {
  return this.mode === this.RS485_HD_MODE;
}
MtsIoSysfs.prototype.execMtsIoSysfs = function(callback) {
  var self = this;
  var command = "/usr/sbin/mts-io-sysfs store ";
  command = command.concat(this.accessoryCard, "/serial-mode ", this.mode);
  console.log("Running command: " + command);
  exec(command, function(error, stdout, stderr) {
  console.log('mts-io-sysfs stdout: ' + stdout);
  console.log('mts-io-sysfs stderr: ' + stderr);
  callback();
//    if (error !== null) {
//      console.error('Error: ' + error);
//    } else {
//      self.execSetRs485(callback);
//    }
  });
}

MtsIoSysfs.prototype.execSetRs485 = function(callback){
//  # enable RS485 half duplex on the uart
//  set-rs485 /dev/ttyAP1 1
//  # disable
//  set-rs485 /dev/ttyAP1 0
  var command = "/usr/bin/set-rs485 " + this.getTtyPath();
  var self = this;
  if (this.isRs485HdMode()) {
      command += " 1";
      console.log("Running command: " + command);
      exec(command, function(error, stdout, stderr) {
        console.log('set-rs485 stdout: ' + stdout);
        console.log('set-rs485 stderr: ' + stderr);
        if (error !== null) {
          console.error('Error: ' + error);
        } else {
          callback();
        }
    });
  } else {
    callback();
  }
}
