/**
 * Copyright 2015 Multitech Inc.
 *
 **/
module.exports = function(RED) {
  "use strict";
  var compatableMcardIDs = ['MTAC-GPIOB'];
  var Gpio = require('mts-gpio');
  var Mcard = require('mts-mcard');

  function setStatus(node, status) {
    switch (status) {
      case "checking": node.status({fill:"grey",shape:"dot",text:"checking mCard"}); break;
      case "connecting": node.status({fill:"grey",shape:"dot",text:"connecting"}); break;
      case "polling": node.status({fill:"blue",shape:"dot",text:"polling"}); break;
      case "incompatable": node.status({fill:"red",shape:"ring",text:"INCOMPATABLE MCARD"}); break;
      case "nocard": node.status({fill:"red",shape:"ring",text:"NO MCARD"}); break;
      default: node.status({}); break;
    }
  }

  function VerifyMcardThenSetupWith(node, config, setupCallback) {
    // For auto detect, check all ports
    if (config.accessoryport == "Auto Detect") {
      Mcard.compatableProductIdPresent(compatableMcardIDs, function(result) {
        node.log('mcard: ' + result);
        if (result == true) {
          setStatus(node, "connecting");
          setupCallback(node, config);
        }
        else {
          setStatus(node, "incompatable");
          node.error('GPIO mCard not found. Compatable mCards: ' + compatableMcardIDs);
        }
      });
    }
    // For a specific port, check the mcard in that port
    else {
      Mcard.getProductId(config.accessoryport, function(result) {
        node.log('mcard: ' + result);
        if (compatableMcardIDs.indexOf(result) != -1) {
          setStatus(node, "connecting");
          setupCallback(node, config);
        }
        else if (result == 'nocard') {
          setStatus(node, "nocard");
          node.error('No mCard in port ' + config.accessoryport + '. Compatable mCards: ' + compatableMcardIDs);
        }
        else {
          setStatus(node, "incompatable");
          node.error('Incompatable mCard in port ' + config.accessoryport + '. Compatable mCards: ' + compatableMcardIDs);
        }
      });
    }
  }


  function GPIODigitalInputNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;
    node.log("Starting GPIO digital input node...");

    setStatus(node, "checking");
    VerifyMcardThenSetupWith(node, config, SetupGPIODigitalInputNode);

    node.log("GPIO digital in node created");
  }

  function SetupGPIODigitalInputNode(node, config) {
      var gpio = new Gpio();
      node.name = config.name;
      node.accessoryport = config.accessoryport;
      node.gpionumber = config.gpionumber;
      node.samplerate = config.samplerate;
      node.sampleinterval = gpio.calculateSampleInterval(node.samplerate);
      node.statefilter = config.statefilter;
      node.timer = null;
      node.statustimer = null;
      node.statustimeout = 500;
      node.previousValue = undefined;

      if (node.statustimeout * 2 > node.sampleinterval) {
         node.statustimeout = 0.5 * node.sampleinterval;
      }

      node.on('close', function() {
          if (node.timer != null) {
              clearInterval(node.timer);
          }
          if (node.statustimer != null) {
              clearTimeout(node.statustimer);
          }
          node.log("digital gpio in close");
      });
      if (node.accessoryport == "Auto Detect") {
          clearInterval(node.timer);
          clearTimeout(node.statustimer);
          node.timer = setInterval(
              function () {
                  setStatus(node, "polling");
                  clearTimeout(node.statustimer);
                  node.statustimer = setTimeout(function() { setStatus(node, ""); }, node.statustimeout);
                  var value = gpio.detectAndReadDigitalValue(node.gpionumber);
                  if (value != null && value != undefined
                  	&& gpio.stateFilterValueChanged(node.statefilter, node.previousValue, value)) {
                      var msg = {payload: value};
                      node.send(msg);
                  }
                  node.previousValue = value;
              },
              node.sampleinterval);
      } else {
          clearInterval(node.timer);
          clearTimeout(node.statustimer);
          node.timer = setInterval(
              function () {
                  setStatus(node, "polling");
                  clearTimeout(node.statustimer);
                  node.statustimer = setTimeout(function() { setStatus(node, ""); }, node.statustimeout);
                  var value = gpio.readDigitalValue(node.accessoryport, node.gpionumber);
                  if (value != null && value != undefined
                  	&& gpio.stateFilterValueChanged(node.statefilter, node.previousValue, value)) {
                      var msg = {payload: value};
                      node.send(msg);
                  }
                  node.previousValue = value;
              },
              node.sampleinterval);
      }
  }

  RED.nodes.registerType("digital gpio in", GPIODigitalInputNode);


  function GPIOAnalogInputNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;
    node.log("Starting GPIO analog input node...");

    setStatus(node, "checking");
    VerifyMcardThenSetupWith(node, config, SetupGPIOAnalogInputNode);

    node.log("GPIO analog in node created");
  }

  function SetupGPIOAnalogInputNode(node, config) {
      var gpio = new Gpio();
      node.name = config.name;
      node.accessoryport = config.accessoryport;
      node.gpionumber = config.gpionumber;
      node.samplerate = config.samplerate;
      node.sampleinterval = gpio.calculateSampleInterval(node.samplerate);
      node.timer = null;
      node.valuerangefilter = config.valuerangefilter;
      node.relationaloperator = config.relationaloperator;
      node.threshold = config.threshold;
      node.statustimer = null;
      node.statustimeout = 500;

      if (node.statustimeout * 2 > node.sampleinterval) {
         node.statustimeout = 0.5 * node.sampleinterval;
      }

      node.on('close', function() {
          if (node.timer != null) {
              clearInterval(node.timer);
          }
          if (node.statustimer != null) {
              clearTimeout(node.statustimer);
          }
          node.log("analog gpio in close");
      });
      if (node.accessoryport == "Auto Detect") {
          clearInterval(node.timer);
          clearTimeout(node.statustimer);
          node.timer = setInterval(
              function () {
                  setStatus(node, "polling");
                  clearTimeout(node.statustimer);
                  node.statustimer = setTimeout(function() { setStatus(node, ""); }, node.statustimeout);
                  var value = gpio.detectAndReadAnalogValue(node.gpionumber);
                  if (value != null && value != undefined
                  	&& gpio.valueRangeFilterPass(node.valuerangefilter, value, node.relationaloperator, node.threshold)) {
                      var msg = {payload: value};
                      node.send(msg);
                  }
                  node.previousValue = value;
              },
              node.sampleinterval);
      } else {
          clearInterval(node.timer);
          clearTimeout(node.statustimer);
          node.timer = setInterval(
              function () {
                  setStatus(node, "polling");
                  clearTimeout(node.statustimer);
                  node.statustimer = setTimeout(function() { setStatus(node, ""); }, node.statustimeout);
                  var value = gpio.readAnalogValue(node.accessoryport, node.gpionumber);
                  if (value != null && value != undefined
                  	&& gpio.valueRangeFilterPass(node.valuerangefilter, value, node.relationaloperator, node.threshold)) {
                      var msg = {payload: value};
                      node.send(msg);
                  }
                  node.previousValue = value;
              },
              node.sampleinterval);
      }
  }

  RED.nodes.registerType("analog gpio in", GPIOAnalogInputNode);


  function GPIODigitalOutputNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;
    node.log("Starting GPIO digital output node...");

    setStatus(node, "checking");
    VerifyMcardThenSetupWith(node, config, SetupGPIODigitalOutputNode);

    node.log("GPIO digital out node created");
  }

  function SetupGPIODigitalOutputNode(node, config) {
      var gpio = new Gpio();
      node.name = config.name;
      node.accessoryport = config.accessoryport;
      node.gpionumber = config.gpionumber;
      setStatus(node, "");

      node.on('input', function(msg) {
          if (node.accessoryport == "Auto Detect") {
             gpio.detectAndWriteDigitalValue(node.gpionumber,
                parseInt(msg.payload));
          } else {
             gpio.writeDigitalValue(node.accessoryport, node.gpionumber,
                parseInt(msg.payload));
          }
      });
  }

  RED.nodes.registerType("digital gpio out", GPIODigitalOutputNode);
}
