module.exports = function(RED) {
  "use strict";
  var compatableMcardIDs = ['MTAC-MFSER-DCE', 'MTAC-MFSER-DTE'];
  var mqtt = require('mqtt');
  var spawn = require('child_process').spawn;
  var https = require('https');
  var Mcard = require('mts-mcard');

  function setStatus(node, status) {
    switch (status) {
      case "checking": node.status({fill:"grey",shape:"dot",text:"checking mCard"}); break;
      case "connecting": node.status({fill:"grey",shape:"dot",text:"connecting"}); break;
      case "connected": node.status({fill:"green",shape:"dot",text:"connected"}); break;
      case "disconnected": node.status({fill:"red",shape:"ring",text:"disconnected"}); break;
      case "incompatable": node.status({fill:"red",shape:"ring",text:"INCOMPATABLE MCARD"}); break;
      case "nocard": node.status({fill:"red",shape:"ring",text:"NO MCARD"}); break;
      default: node.status({}); break;
    }
  }

  function VerifyMcardThenSetupWith(node, config, setupCallback) {
    Mcard.getProductId(config.accessoryport, function(result) {
      node.log('mcard: ' + result);
      if (compatableMcardIDs.indexOf(result) != -1) {
        setStatus(node, "connecting");
        setupCallback(node, config);
      }
      else if (result == 'nocard') {
        setStatus(node, "nocard");
        node.error('No mCard in port ' + config.accessoryport + '. Compatable mCards: ' + compatableMcardIDs);
      }
      else {
        setStatus(node, "incompatable");
        node.error('Incompatable mCard in port ' + config.accessoryport + '. Compatable mCards: ' + compatableMcardIDs);
      }
    });
  }

  function Ensure110CompatibleConfig(node, config) {
    //Update flowcontrol config from 1.0.33 to 1.1.0 if needed
    //NOTE: This doesn't save the config. It only updates it for the current runtime
    if (config.xonxoff || config.dsrdtr || config.rtscts) {
        node.log("updating config to be compatable with 1.1.0");
        if (config.rtscts == "true" || config.dsrdtr == "true") {
            config.flowcontrol = "hardware";
        }
        else if (config.xonxoff == "true") {
            config.flowcontrol = "software";
        }
        else {
            config.flowcontrol = "disabled";
        }
    }
  }

  function MultiSerialIn(config) {
    RED.nodes.createNode(this, config);
    var node = this;

    setStatus(node, "checking");
    VerifyMcardThenSetupWith(node, config, SetupMultiSerialIn);

    node.log("multi serial in node created");
  }

  function SetupMultiSerialIn(node, config) {
    Ensure110CompatibleConfig(node, config);

    node.bytecount = config.bytecount || 'None';

    var params = [ "--" + config.datatype,
              '--baudrate', config.serialbaud,
              '--bytesize', config.databits,
              '--parity', config.parity,
              '--stopbits', config.stopbits,
              '--mode', config.mode,
              '--accessoryPort', config.accessoryport];
    if (config.datatype == "readbytes") {
      params.push('--readNumBytes');
      params.push(node.bytecount);
    }
    if (config.timeout) {
      params.push('--timeout');
      params.push(config.timeout);
    }

    if (config.flowcontrol == "software") {
      params.push('--xonxoff');
    }
    else if (config.flowcontrol == "hardware") {
      params.push('--rtscts');
    }
    node.log(params.toString());
    node.serialstreamer = spawn('/opt/serial-streamer/serial_streamer', params);
    node.serialstreamer.stdout.on('data', function (data) {
      //Remove line endings
      data = data.toString().trim().replace(/^.*(?=aep-utils)/g, '');
      node.log('ss stdout: ' + data);
      if (data.match(/Connected/g)) {
          setStatus(node, "connected");
      }
    });

    node.serialstreamer.stderr.on('data', function (data) {
      //Remove line endings
      data = data.toString().trim().replace(/^.*(?=aep-utils)/g, '');
      node.log('ss stderr: ' + data);
      if (data.match(/Connected/g)) {
        setStatus(node, "connected");
      }
    });

    node.serialstreamer.on('close', function (code) {
      node.log('Serialstreamer exited with code ' + code);
      setStatus(node, "disconnected");
    });
    var client;
    if (config.datatype == "readbytes") {
      client = mqtt.connect('mqtt://127.0.0.1', {encoding: 'binary'});
    } else if (config.datatype == "readlines") {
      client = mqtt.connect('mqtt://127.0.0.1');
    }
    node.log('SerialStreamer node subscribing to ' + 'serial/ap' + config.accessoryport + '/receive/' + node.serialstreamer.pid);
    client.subscribe('serial/ap' + config.accessoryport + '/receive/' + node.serialstreamer.pid, {'qos': 1});

    client.on('message', function(topic, message) {
      //node.log("Received message: " + message.toString());
      node.send({payload: message});
    });
    node.on("close", function() {
      node.log("multi serial in close");
      client.end();
      node.serialstreamer.kill();
      setStatus(node, "disconnected");
    });
  }

  RED.nodes.registerType("multi serial in", MultiSerialIn);


  function MultiSerialOut(config) {
    RED.nodes.createNode(this, config);
    var node = this;

    setStatus(node, "checking");
    VerifyMcardThenSetupWith(node, config, SetupMultiSerialOut);

    node.log("multi serial out node created");
  }

  function SetupMultiSerialOut(node, config) {
    Ensure110CompatibleConfig(node, config);

    var params = [ '--write',
            '--baudrate', config.serialbaud,
            '--bytesize', config.databits,
            '--parity', config.parity,
            '--stopbits', config.stopbits,
            '--mode', config.mode,
            '--accessoryPort', config.accessoryport];

    if (config.flowcontrol == "software") {
      params.push('--xonxoff');
    }
    else if (config.flowcontrol == "hardware") {
      params.push('--rtscts');
    }

    node.log(params.toString());
    node.serialstreamer = spawn('/opt/serial-streamer/serial_streamer', params);
    node.serialstreamer.stdout.on('data', function (data) {
      //Remove line endings
      data = data.toString().trim().replace(/^.*(?=aep-utils)/g, '');
      node.log('ss stdout: ' + data);
      if (data.match(/Connected/g)) {
          setStatus(node, "connected");
      }
    });

    node.serialstreamer.stderr.on('data', function (data) {
      //Remove line endings
      data = data.toString().trim().replace(/^.*(?=aep-utils)/g, '');
      node.log('ss stderr: ' + data);
      if (data.match(/Connected/g)) {
        setStatus(node, "connected");
      }
    });

    node.serialstreamer.on('close', function (code) {
      node.log('Serialstreamer exited with code ' + code);
      setStatus(node, "disconnected");
    });
    node.log('Spawned serial-streamer pid: ' + node.serialstreamer.pid);
    var client = mqtt.connect('mqtt://127.0.0.1');
    node.on('input', function(msg) {
      //node.log("Sending MQTT message: " + msg.toString());
      client.publish('serial/ap'
          + config.accessoryport + '/transmit/' + node.serialstreamer.pid, msg.payload);
    });
    node.on("close", function() {
      node.log("multi serial out on close");
      client.end();
      node.serialstreamer.kill();
      if (!node.serialstreamer.connected) {
        setStatus(node, "disconnected");
      }
    });
  }

  RED.nodes.registerType("multi serial out", MultiSerialOut);
};
