
module.exports = function(RED) {
  "use strict";
  var compatableRadioCodes = ['H5', 'LAT1', 'LEU1', 'LVW2', 'LNA3'];
  var http = require('http');
  var pollInterval = 10000;

  function setStatus(node, status) {
    switch (status) {
      case "polling": node.status({fill:"blue",shape:"dot",text:"polling"}); break;
      case "sending": node.status({fill:"blue",shape:"dot",text:"sending"}); break;
      case "error-poll": node.status({fill:"red",shape:"ring",text:"ERROR POLLING"}); break;
      case "error-delete": node.status({fill:"red",shape:"ring",text:"ERROR DELETING SMS"}); break;
      case "error-send": node.status({fill:"red",shape:"ring",text:"ERROR SENDING"}); break;
      case "invalidradio": node.status({fill:"red",shape:"ring",text:"INVALID RADIO"}); break;
      default: node.status({}); break;
    }
  }

  function VerifyRadioThenSetupWith(node, config, setupCallback) {
    http.get('http://127.0.0.1/api/system/radio', function(result) {
      result.on('data', function(data) {
        node.log(data);
        var radio = JSON.parse(data);

        if (compatableRadioCodes.indexOf(radio.result.code) == -1) {
          setStatus(node, "invalidradio");
          node.error('Compatable cellular radio not found. Compatable radios: ' + compatableRadioCodes);
          return;
        }

        setupCallback(node, config);
      });
    }).on('error', function(e) {
      node.error(e);
      callback('error');
    });
  }

  function pollSms(node) {
    setStatus(node, "polling");
    clearTimeout(node.statustimer);
    node.statustimer = setTimeout(function() { setStatus(node, ""); }, node.statustimeout);
    var getRequest = http.request({hostname: '127.0.0.1', port: 80, path: '/api/sms/inbox', method: 'GET'}, function(getResponse) {
      var responseBody = '';
      getResponse.on('data', function(data) {
        responseBody += data;
      });
      getResponse.on('end', function() {
        var response = JSON.parse(responseBody);
        var array = response["result"];
        array.forEach(function(object) {
          node.send({"payload": object["message"], "guid": object["guid"], "source": object["source"], "timestamp": object["timestamp"]});
          var guid = object["guid"];
          node.log("Received SMS GUID: " + guid)
          var deleteRequest = http.request({hostname: '127.0.0.1', port: 80, path: '/api/sms/inbox/' + guid, method: 'DELETE'}
          ).on('error', function () {
            clearTimeout(node.statustimer);
            setStatus(node, "error-delete");
            node.error(e);
          }).on('end', function() {
            node.log("Removed SMS from queue with GUID: " + guid);
          });
          deleteRequest.end();
        });
      });
    }).on('error', function(e) {
        node.error(e);
        clearTimeout(node.statustimer);
        setStatus(node, "error-poll");
    });
    getRequest.end();
  }

  function SmsInNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;

    VerifyRadioThenSetupWith(node, config, SetupSmsInNode);

    node.log("sms in node created");
  }

  function SetupSmsInNode(node, config) {
    node.timer = null;
    node.statustimer = null;
    node.statustimeout = 1000;

    node.on('close', function() {
      if (node.timer != null) {
          clearInterval(node.timer);
      }
      if (node.statustimer != null) {
          clearTimeout(node.statustimer);
      }
      node.log("sms in close");
    });

    node.timer = setInterval(pollSms, pollInterval, node);
  }

  RED.nodes.registerType("sms in", SmsInNode);


  function SmsOutNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;

    VerifyRadioThenSetupWith(node, config, SetupSmsOutNode);

    node.log("sms out node created");
  }

  function SetupSmsOutNode(node, config) {
    node.message = config.message;

    if (config.recipients) {
      node.recipients = config.recipients.split(",");
    }
    else {
      node.recipients = "";
    }

    for (var i = 0; i < node.recipients.length; i++) {
      node.recipients[i] = node.recipients[i].replace("-", "").replace("(", "").replace(")", "").replace(" ", "").trim();
    }
    node.on('input', function(msg) {
      var message = msg.payload || node.message;
      message = message.toString();
      var recipients = msg.recipients || node.recipients;
      var postData = JSON.stringify({ message : message,
        recipients : recipients});
      var options = {
        hostname: '127.0.0.1',
        path: '/api/sms/outbox',
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
          'Content-Length': Buffer.byteLength(postData, 'utf8')
        }
      };

      setStatus(node, "sending");
      var postRequest = http.request(options, function(postResponse) {
        var responseBody = '';
        postResponse.on('data', function(data) {
          responseBody += data;
        });
        postResponse.on('end', function() {
          var object = JSON.parse(responseBody);
          if (object.code == 200) {
            node.log("Sent SMS message to: " + node.recipients.toString() + ", Response: " + object.status);
            setStatus(node, "");
          } else {
            node.error("SMS error: " + object.error);
            setStatus(node, "error-send");
          }
        })
      }).on('error', function(e) {
        setStatus(node, "error-send");
        console.log('Problem with SMS send: ' + e.message);
      });

      postRequest.write(postData);
      postRequest.end();
    });
  }

  RED.nodes.registerType("sms out", SmsOutNode);

};
